package com.simibubi.create.block;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllItems;
import com.simibubi.create.AllTileEntities;
import com.simibubi.create.item.ItemBlueprint;
import com.simibubi.create.schematic.Cuboid;
import com.simibubi.create.schematic.MaterialChecklist;
import com.simibubi.create.schematic.SchematicWorld;
import com.simibubi.create.utility.TileEntitySynced;

import net.minecraft.block.BlockState;
import net.minecraft.block.Blocks;
import net.minecraft.entity.player.PlayerEntity;
import net.minecraft.entity.player.PlayerInventory;
import net.minecraft.inventory.container.Container;
import net.minecraft.inventory.container.INamedContainerProvider;
import net.minecraft.item.BlockItem;
import net.minecraft.item.ItemStack;
import net.minecraft.item.Items;
import net.minecraft.nbt.CompoundNBT;
import net.minecraft.nbt.ListNBT;
import net.minecraft.nbt.NBTUtil;
import net.minecraft.network.PacketBuffer;
import net.minecraft.tileentity.ITickableTileEntity;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.tileentity.TileEntityType;
import net.minecraft.util.Direction;
import net.minecraft.util.SoundCategory;
import net.minecraft.util.SoundEvents;
import net.minecraft.util.math.AxisAlignedBB;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.MathHelper;
import net.minecraft.util.text.ITextComponent;
import net.minecraft.util.text.StringTextComponent;
import net.minecraft.world.gen.feature.template.Template;
import net.minecraftforge.api.distmarker.Dist;
import net.minecraftforge.api.distmarker.OnlyIn;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.items.CapabilityItemHandler;
import net.minecraftforge.items.IItemHandler;
import net.minecraftforge.items.ItemStackHandler;

public class SchematicannonTileEntity extends TileEntitySynced implements ITickableTileEntity, INamedContainerProvider {

	public static final int NEIGHBOUR_CHECKING = 100;
	public static final int PLACEMENT_DELAY = 10;
	public static final float FUEL_PER_GUNPOWDER = .2f;
	public static final float FUEL_USAGE_RATE = .0005f;
	public static final int SKIPS_PER_TICK = 10;
	public static final int MAX_ANCHOR_DISTANCE = 256;

	public enum State {
		STOPPED, PAUSED, RUNNING;
	}

	// Inventory
	public SchematicannonInventory inventory;

	// Sync
	public boolean sendUpdate;
	public boolean dontUpdateChecklist;
	public int neighbourCheckCooldown;

	// Printer
	private SchematicWorld blockReader;
	public BlockPos currentPos;
	public BlockPos schematicAnchor;
	public boolean schematicLoaded;
	public boolean missingBlock;
	public boolean blockNotLoaded;
	public boolean hasCreativeCrate;
	private int printerCooldown;
	private int skipsLeft;
	private boolean blockSkipped;

	public BlockPos target;
	public BlockPos previousTarget;
	public List<IItemHandler> attachedInventories;
	public List<LaunchedBlock> flyingBlocks;
	public MaterialChecklist checklist;

	// Gui information
	public float fuelLevel;
	public float bookPrintingProgress;
	public float schematicProgress;
	public String statusMsg;
	public State state;
	public int blocksPlaced;
	public int blocksToPlace;

	// Settings
	public int replaceMode;
	public boolean skipMissing;
	public boolean replaceTileEntities;

	// Render
	public boolean firstRenderTick;

	public class SchematicannonInventory extends ItemStackHandler {
		public SchematicannonInventory() {
			super(5);
		}

		@Override
		protected void onContentsChanged(int slot) {
			super.onContentsChanged(slot);
			markDirty();
		}

		@Override
		public boolean isItemValid(int slot, ItemStack stack) {
			switch (slot) {
			case 0: // Blueprint Slot
				return AllItems.BLUEPRINT.typeOf(stack);
			case 1: // Blueprint output
				return false;
			case 2: // Book input
				return stack.getItem() == Items.BOOK || stack.getItem() == Items.WRITTEN_BOOK;
			case 3: // Material List output
				return false;
			case 4: // Gunpowder
				return stack.getItem() == Items.GUNPOWDER;
			default:
				return super.isItemValid(slot, stack);
			}
		}
	}

	public class LaunchedBlock {
		public int totalTicks;
		public int ticksRemaining;
		public BlockPos target;
		public BlockState state;

		public LaunchedBlock(BlockPos target, BlockState state) {
			this.target = target;
			this.state = state;
			totalTicks = (int) (Math.max(10, MathHelper.sqrt(MathHelper.sqrt(target.distanceSq(pos))) * 4f));
			ticksRemaining = totalTicks;
		}

		public LaunchedBlock(BlockPos target, BlockState state, int ticksLeft, int total) {
			this.target = target;
			this.state = state;
			this.totalTicks = total;
			this.ticksRemaining = ticksLeft;
		}

		public void update() {
			if (ticksRemaining > 0)
				ticksRemaining--;
		}
	}

	public SchematicannonTileEntity() {
		this(AllTileEntities.Schematicannon.type);
	}

	@Override
	public double getMaxRenderDistanceSquared() {
		return 65536.0D;
	}

	@OnlyIn(Dist.CLIENT)
	@Override
	public AxisAlignedBB getRenderBoundingBox() {
		return INFINITE_EXTENT_AABB;
	}

	public SchematicannonTileEntity(TileEntityType<?> tileEntityTypeIn) {
		super(tileEntityTypeIn);
		attachedInventories = new LinkedList<>();
		flyingBlocks = new LinkedList<>();
		inventory = new SchematicannonInventory();
		statusMsg = "Idle";
		state = State.STOPPED;
		replaceMode = 2;
		neighbourCheckCooldown = NEIGHBOUR_CHECKING;
		checklist = new MaterialChecklist();
	}

	public void findInventories() {
		hasCreativeCrate = false;
		attachedInventories.clear();
		for (Direction facing : Direction.values()) {

			if (!world.isBlockPresent(pos.offset(facing)))
				continue;

			if (AllBlocks.CREATIVE_CRATE.typeOf(world.getBlockState(pos.offset(facing)))) {
				hasCreativeCrate = true;
			}

			TileEntity tileEntity = world.getTileEntity(pos.offset(facing));
			if (tileEntity != null) {
				LazyOptional<IItemHandler> capability = tileEntity
						.getCapability(CapabilityItemHandler.ITEM_HANDLER_CAPABILITY, facing.getOpposite());
				if (capability.isPresent()) {
					attachedInventories.add(capability.orElse(null));
				}
			}
		}
	}

	@Override
	public void read(CompoundNBT compound) {
		inventory.deserializeNBT(compound.getCompound("Inventory"));

		if (compound.contains("Running"))
			currentPos = NBTUtil.readBlockPos(compound.getCompound("CurrentPos"));

		readClientUpdate(compound);
		super.read(compound);
	}

	@Override
	public void readClientUpdate(CompoundNBT compound) {

		// Gui information
		statusMsg = compound.getString("Status");
		schematicProgress = compound.getFloat("Progress");
		bookPrintingProgress = compound.getFloat("PaperProgress");
		fuelLevel = compound.getFloat("Fuel");
		state = State.valueOf(compound.getString("State"));
		blocksPlaced = compound.getInt("AmountPlaced");
		blocksToPlace = compound.getInt("AmountToPlace");

		// Settings
		CompoundNBT options = compound.getCompound("Options");
		replaceMode = options.getInt("ReplaceMode");
		skipMissing = options.getBoolean("SkipMissing");
		replaceTileEntities = options.getBoolean("ReplaceTileEntities");

		// Printer & Flying Blocks
		if (compound.contains("Target"))
			target = NBTUtil.readBlockPos(compound.getCompound("Target"));
		if (compound.contains("FlyingBlocks"))
			readFlyingBlocks(compound);

	}

	protected void readFlyingBlocks(CompoundNBT compound) {
		ListNBT tagBlocks = compound.getList("FlyingBlocks", 10);
		if (tagBlocks.isEmpty())
			flyingBlocks.clear();

		boolean pastDead = false;

		for (int i = 0; i < tagBlocks.size(); i++) {
			CompoundNBT c = tagBlocks.getCompound(i);

			BlockPos readBlockPos = NBTUtil.readBlockPos(c.getCompound("Target"));
			BlockState readBlockState = NBTUtil.readBlockState(c.getCompound("Block"));
			int int1 = c.getInt("TicksLeft");
			int int2 = c.getInt("TotalTicks");

			// Always write to Server tile
			if (world == null || !world.isRemote) {
				flyingBlocks.add(new LaunchedBlock(readBlockPos, readBlockState, int1, int2));
				continue;
			}

			// Delete all Client side blocks that are now missing on the server
			while (!pastDead && !flyingBlocks.isEmpty() && !flyingBlocks.get(0).target.equals(readBlockPos)) {
				flyingBlocks.remove(0);
			}

			pastDead = true;

			// Add new server side blocks
			if (i >= flyingBlocks.size()) {
				flyingBlocks.add(new LaunchedBlock(readBlockPos, readBlockState, int1, int2));
				continue;
			}

			// Don't do anything with existing
		}
	}

	@Override
	public CompoundNBT write(CompoundNBT compound) {
		compound.put("Inventory", inventory.serializeNBT());

		if (state == State.RUNNING) {
			compound.putBoolean("Running", true);
			compound.put("CurrentPos", NBTUtil.writeBlockPos(currentPos));
		}

		writeToClient(compound);
		return super.write(compound);
	}

	@Override
	public CompoundNBT writeToClient(CompoundNBT compound) {

		// Gui information
		compound.putFloat("Progress", schematicProgress);
		compound.putFloat("PaperProgress", bookPrintingProgress);
		compound.putFloat("Fuel", fuelLevel);
		compound.putString("Status", statusMsg);
		compound.putString("State", state.name());
		compound.putInt("AmountPlaced", blocksPlaced);
		compound.putInt("AmountToPlace", blocksToPlace);

		// Settings
		CompoundNBT options = new CompoundNBT();
		options.putInt("ReplaceMode", replaceMode);
		options.putBoolean("SkipMissing", skipMissing);
		options.putBoolean("ReplaceTileEntities", replaceTileEntities);
		compound.put("Options", options);

		// Printer & Flying Blocks
		if (target != null)
			compound.put("Target", NBTUtil.writeBlockPos(target));
		ListNBT tagBlocks = new ListNBT();
		for (LaunchedBlock b : flyingBlocks) {
			CompoundNBT c = new CompoundNBT();
			c.putInt("TotalTicks", b.totalTicks);
			c.putInt("TicksLeft", b.ticksRemaining);
			c.put("Target", NBTUtil.writeBlockPos(b.target));
			c.put("Block", NBTUtil.writeBlockState(b.state));
			tagBlocks.add(c);
		}
		compound.put("FlyingBlocks", tagBlocks);

		return compound;
	}

	@Override
	public void tick() {
		if (neighbourCheckCooldown-- <= 0) {
			neighbourCheckCooldown = NEIGHBOUR_CHECKING;
			findInventories();
		}

		firstRenderTick = true;
		previousTarget = target;
		tickFlyingBlocks();

		if (world.isRemote)
			return;

		// Update Fuel and Paper
		tickPaperPrinter();
		refillFuelIfPossible();

		// Update Printer
		skipsLeft = SKIPS_PER_TICK;
		blockSkipped = true;

		while (blockSkipped && skipsLeft-- > 0)
			tickPrinter();

		schematicProgress = 0;
		if (blocksToPlace > 0)
			schematicProgress = (float) blocksPlaced / blocksToPlace;

		// Update Client Tile
		if (sendUpdate) {
			sendUpdate = false;
			world.notifyBlockUpdate(pos, getBlockState(), getBlockState(), 6);
		}
	}

	protected void tickPrinter() {
		ItemStack blueprint = inventory.getStackInSlot(0);
		blockSkipped = false;

		// Skip if not Active
		if (state == State.STOPPED) {
			if (schematicLoaded)
				resetPrinter();
			return;
		}

		if (blueprint.isEmpty()) {
			state = State.STOPPED;
			statusMsg = "Idle";
			sendUpdate = true;
			return;
		}

		if (state == State.PAUSED && !blockNotLoaded && !missingBlock && fuelLevel > FUEL_USAGE_RATE)
			return;

		// Initialize Printer
		if (!schematicLoaded) {
			initializePrinter(blueprint);
			return;
		}

		// Cooldown from last shot
		if (printerCooldown > 0) {
			printerCooldown--;
			return;
		}

		// Check Fuel
		if (fuelLevel <= 0) {
			fuelLevel = 0;
			state = State.PAUSED;
			statusMsg = "Out of Gunpowder";
			sendUpdate = true;
			return;
		}

		// Update Target
		if (hasCreativeCrate) {
			if (missingBlock) {
				missingBlock = false;
				state = State.RUNNING;
			}
		}
		if (!missingBlock && !blockNotLoaded) {
			advanceCurrentPos();

			// End reached
			if (state == State.STOPPED)
				return;

			sendUpdate = true;
			target = schematicAnchor.add(currentPos);
		}

		// Check block
		if (!getWorld().isAreaLoaded(target, 0)) {
			blockNotLoaded = true;
			statusMsg = "Block is not loaded";
			state = State.PAUSED;
			return;
		} else {
			if (blockNotLoaded) {
				blockNotLoaded = false;
				state = State.RUNNING;
			}
		}
		
		BlockState blockState = blockReader.getBlockState(target);
		if (!shouldPlace(target, blockState)) {
			statusMsg = "Searching";
			blockSkipped = true;
			return;
		}

		// Find Item
		ItemStack requiredItem = getItemForBlock(blockState);
		if (!findItemInAttachedInventories(requiredItem)) {
			if (skipMissing) {
				statusMsg = "Skipping";
				blockSkipped = true;
				if (missingBlock) {
					missingBlock = false;
					state = State.RUNNING;
				}
				return;
			}

			missingBlock = true;
			state = State.PAUSED;
			statusMsg = "Missing " + blockState.getBlock().getNameTextComponent().getFormattedText();
			return;
		}

		// Success
		state = State.RUNNING;
		if (blockState.getBlock() != Blocks.AIR)
			statusMsg = "Placing: " + blocksPlaced + " / " + blocksToPlace;
		else
			statusMsg = "Clearing Blocks";
		launchBlock(target, blockState);
		printerCooldown = PLACEMENT_DELAY;
		fuelLevel -= FUEL_USAGE_RATE;
		sendUpdate = true;
		missingBlock = false;
	}

	protected void initializePrinter(ItemStack blueprint) {
		if (!blueprint.hasTag()) {
			state = State.STOPPED;
			statusMsg = "Invalid Blueprint";
			sendUpdate = true;
			return;
		}

		if (!blueprint.getTag().getBoolean("Deployed")) {
			state = State.STOPPED;
			statusMsg = "Blueprint not Deployed";
			sendUpdate = true;
			return;
		}

		// Load blocks into reader
		Template activeTemplate = ItemBlueprint.getSchematic(blueprint);
		BlockPos anchor = NBTUtil.readBlockPos(blueprint.getTag().getCompound("Anchor"));
		
		if (activeTemplate.getSize().equals(BlockPos.ZERO)) {
			state = State.STOPPED;
			statusMsg = "Schematic File Expired";
			inventory.setStackInSlot(0, ItemStack.EMPTY);
			inventory.setStackInSlot(1, new ItemStack(AllItems.EMPTY_BLUEPRINT.get()));
			return;
		}
		
		if (!anchor.withinDistance(getPos(), MAX_ANCHOR_DISTANCE)) {
			state = State.STOPPED;
			statusMsg = "Target too Far Away";
			return;
		}
		
		schematicAnchor = anchor;
		blockReader = new SchematicWorld(new HashMap<>(), new Cuboid(), schematicAnchor);
		activeTemplate.addBlocksToWorld(blockReader, schematicAnchor, ItemBlueprint.getSettings(blueprint));
		schematicLoaded = true;
		state = State.PAUSED;
		statusMsg = "Ready";
		updateChecklist();
		sendUpdate = true;
		blocksToPlace += blocksPlaced;
		currentPos = currentPos != null ? currentPos.west() : blockReader.getBounds().getOrigin().west();
	}

	protected ItemStack getItemForBlock(BlockState blockState) {
		return new ItemStack(BlockItem.BLOCK_TO_ITEM.getOrDefault(blockState.getBlock(), Items.AIR));
	}

	protected boolean findItemInAttachedInventories(ItemStack requiredItem) {
		if (hasCreativeCrate)
			return true;

		for (IItemHandler iItemHandler : attachedInventories) {
			for (int slot = 0; slot < iItemHandler.getSlots(); slot++) {
				ItemStack stackInSlot = iItemHandler.getStackInSlot(slot);
				if (!stackInSlot.isItemEqual(requiredItem))
					continue;
				if (!iItemHandler.extractItem(slot, 1, false).isEmpty())
					return true;
			}
		}
		return false;
	}

	protected void advanceCurrentPos() {
		BlockPos size = blockReader.getBounds().getSize();
		currentPos = currentPos.offset(Direction.EAST);
		BlockPos posInBounds = currentPos.subtract(blockReader.getBounds().getOrigin());

		if (posInBounds.getX() > size.getX())
			currentPos = new BlockPos(blockReader.getBounds().x, currentPos.getY(), currentPos.getZ() + 1).west();
		if (posInBounds.getZ() > size.getZ())
			currentPos = new BlockPos(currentPos.getX(), currentPos.getY() + 1, blockReader.getBounds().z).west();

		// End reached
		if (currentPos.getY() > size.getY()) {
			inventory.setStackInSlot(0, ItemStack.EMPTY);
			inventory.setStackInSlot(1,
					new ItemStack(AllItems.EMPTY_BLUEPRINT.get(), inventory.getStackInSlot(1).getCount() + 1));
			state = State.STOPPED;
			statusMsg = "Finished";
			resetPrinter();
			target = getPos().add(1, 0, 0);
			world.playSound(null, pos.getX(), pos.getY(), pos.getZ(), SoundEvents.BLOCK_NOTE_BLOCK_BELL,
					SoundCategory.BLOCKS, 1, .7f);
			sendUpdate = true;
			return;
		}
	}

	protected void resetPrinter() {
		schematicLoaded = false;
		schematicAnchor = null;
		currentPos = null;
		blockReader = null;
		missingBlock = false;
		sendUpdate = true;
		schematicProgress = 0;
		blocksPlaced = 0;
		blocksToPlace = 0;
	}

	protected boolean shouldPlace(BlockPos pos, BlockState state) {
		BlockState toReplace = world.getBlockState(pos);
		boolean placingAir = state.getBlock() == Blocks.AIR;

		if (toReplace.getBlockState() == state)
			return false;
		if (pos.withinDistance(getPos(), 2f))
			return false;
		if (!replaceTileEntities && toReplace.hasTileEntity())
			return false;
		
		// Block doesnt have a mapping (Water, lava, etc)
		if (getItemForBlock(state).getItem() == Items.AIR && state.getBlock() != Blocks.AIR)
			return false;

		if (replaceMode == 3)
			return true;
		if (replaceMode == 2 && !placingAir)
			return true;
		if (replaceMode == 1 && (state.isNormalCube(blockReader, pos.subtract(schematicAnchor))
				|| !toReplace.isNormalCube(world, pos)) && !placingAir)
			return true;
		if (replaceMode == 0 && !toReplace.isNormalCube(world, pos) && !placingAir)
			return true;

		return false;
	}

	protected void tickFlyingBlocks() {
		List<LaunchedBlock> toRemove = new LinkedList<>();
		for (LaunchedBlock b : flyingBlocks) {
			b.update();
			if (b.ticksRemaining <= 0 && !world.isRemote) {
				world.setBlockState(b.target, b.state, 18);
				toRemove.add(b);
			}
		}
		flyingBlocks.removeAll(toRemove);
	}

	protected void refillFuelIfPossible() {
		if (1 - fuelLevel + 1 / 128f < FUEL_PER_GUNPOWDER)
			return;
		if (inventory.getStackInSlot(4).isEmpty())
			return;

		inventory.getStackInSlot(4).shrink(1);
		fuelLevel += FUEL_PER_GUNPOWDER;
		sendUpdate = true;
	}

	protected void tickPaperPrinter() {
		int BookInput = 2;
		int BookOutput = 3;

		ItemStack blueprint = inventory.getStackInSlot(0);
		ItemStack paper = inventory.extractItem(BookInput, 1, true);
		boolean outputFull = inventory.getStackInSlot(BookOutput).getCount() == inventory.getSlotLimit(BookOutput);

		if (paper.isEmpty() || outputFull) {
			if (bookPrintingProgress != 0)
				sendUpdate = true;
			bookPrintingProgress = 0;
			dontUpdateChecklist = false;
			return;
		}

		if (!schematicLoaded) {
			if (!blueprint.isEmpty())
				initializePrinter(blueprint);
			return;
		}

		if (bookPrintingProgress >= 1) {
			bookPrintingProgress = 0;

			if (!dontUpdateChecklist)
				updateChecklist();

			dontUpdateChecklist = true;
			inventory.extractItem(BookInput, 1, false);
			ItemStack stack = checklist.createItem();
			stack.setCount(inventory.getStackInSlot(BookOutput).getCount() + 1);
			inventory.setStackInSlot(BookOutput, stack);
			sendUpdate = true;
			return;
		}

		bookPrintingProgress += 0.05f;
		sendUpdate = true;
	}

	protected void launchBlock(BlockPos target, BlockState state) {
		if (state.getBlock() != Blocks.AIR)
			blocksPlaced++;
		flyingBlocks.add(new LaunchedBlock(target, state));
		world.playSound(null, pos.getX(), pos.getY(), pos.getZ(), SoundEvents.ENTITY_GENERIC_EXPLODE, SoundCategory.BLOCKS,
				.1f, 1.1f);
	}

	public void sendToContainer(PacketBuffer buffer) {
		buffer.writeBlockPos(getPos());
		buffer.writeCompoundTag(getUpdateTag());
	}

	@Override
	public Container createMenu(int id, PlayerInventory inv, PlayerEntity player) {
		return new SchematicannonContainer(id, inv, this);
	}

	@Override
	public ITextComponent getDisplayName() {
		return new StringTextComponent(getType().getRegistryName().toString());
	}

	public void updateChecklist() {
		checklist.required.clear();
		checklist.blocksNotLoaded = false;
		
		if (schematicLoaded) {
			blocksToPlace = blocksPlaced;
			for (BlockPos pos : blockReader.getAllPositions()) {
				BlockState required = blockReader.getBlockState(pos.add(schematicAnchor));
				
				if (!getWorld().isAreaLoaded(pos.add(schematicAnchor), 0)) {
					checklist.warnBlockNotLoaded();
					continue;
				}
				if (!shouldPlace(pos.add(schematicAnchor), required))
					continue;
				ItemStack requiredItem = getItemForBlock(required);
				if (requiredItem.isEmpty())
					continue;
				checklist.require(requiredItem.getItem());
				blocksToPlace++;
			}
		}
		checklist.gathered.clear();
		for (IItemHandler inventory : attachedInventories) {
			for (int slot = 0; slot < inventory.getSlots(); slot++) {
				ItemStack stackInSlot = inventory.getStackInSlot(slot);
				if (inventory.extractItem(slot, 1, true).isEmpty())
					continue;
				checklist.collect(stackInSlot);
			}
		}
		sendUpdate = true;
	}

}
